from collections.abc import Callable
from dataclasses import dataclass
from functools import wraps
from inspect import Signature, signature
from typing import Any

from decorative_secrets._utilities import (
    get_function_signature_applicable_args_kwargs,
    get_signature_parameter_names_defaults,
    merge_function_signature_args_kwargs,
    unwrap_function,
)
from decorative_secrets.utilities import iscoroutinefunction

__all__: tuple[str, ...] = (
    "ApplyConditionalDefaultsOptions",
    "apply_conditional_defaults",
)


def apply_conditional_defaults(
    condition: Callable[..., bool], *default_args: Any, **default_kwargs: Any
) -> Callable[..., Callable[..., Any]]:
    """
    This function decorates another function in order to apply a set of
    *default* keyword or positional/keyword argument values dependent on the
    outcome of passing an applicable subset of those arguments to the
    `condition` function.

    Parameters:
        condition: A function which accepts a subset of the decorated
            function's arguments and returns a boolean value indicating
            whether or not to apply the default argument values.
        *default_args: A set of positional argument values to apply as
            defaults if the condition is met. Note: If any of these arguments
            are an instance of [ApplyConditionalDefaultsOptions
            ](./#decorative_secrets.defaults.ApplyConditionalDefaultsOptions),
            the options will be used to determine decorator behavior, and will
            be redacted from the arguments passed to the decorated function.
        **default_kwargs: A set of keyword argument values to apply as
            defaults if the condition is met.

    Examples:
        ```python
        from decorative_secrets.defaults import apply_conditional_defaults


        @apply_conditional_defaults(
            lambda environment: environment == "prod",
            source_directory="/in/prod",
            target_directory="/out/prod",
        )
        @apply_conditional_defaults(
            lambda environment: environment == "dev",
            source_directory="/in/dev",
            target_directory="/out/dev",
        )
        @apply_conditional_defaults(
            lambda environment: environment == "stage",
            source_directory="/in/stage",
            target_directory="/out/stage",
        )
        def get_environment_source_target(
            environment: str = "dev",
            source_directory: str = "/dev/null",
            target_directory: str = "/dev/null",
        ) -> tuple[str, str, str]:
            return (environment, source_directory, target_directory)


        get_environment_source_target("stage")
        # ('stage', '/in/stage', '/out/stage')

        get_environment_source_target(environment="prod")
        # ('prod', '/in/prod', '/out/prod')

        get_environment_source_target()
        # ('dev', '/in/dev', '/out/dev')
        ```
    """
    options: ApplyConditionalDefaultsOptions
    default_args, options = _get_args_options(*default_args)

    def decorating_function(
        function: Callable[..., Any],
    ) -> Callable[..., Any]:
        original_function: Callable[..., Any] = unwrap_function(function)
        function_signature: Signature = signature(original_function)
        condition_signature: Signature = signature(condition)

        def get_args_kwargs(
            *args: Any, **kwargs: Any
        ) -> tuple[tuple[Any, ...], dict[str, Any]]:
            # First we consolidate the keyword arguments with any arguments
            # which are passed to parameters which can be either positional
            # *or* keyword arguments, and were passed as positional arguments
            args = merge_function_signature_args_kwargs(
                function_signature, args, kwargs
            )
            # Get the arguments and keyword arguments applicable to the
            # condition function
            condition_args: tuple[Any, ...]
            condition_kwargs: dict[str, Any]
            kwargs_or_defaults: dict[str, Any] = kwargs.copy()
            # Use function signature defaults for any missing keyword
            # parameters
            key: str
            value: Any
            for key, value in get_signature_parameter_names_defaults(
                function_signature
            ).items():
                if (
                    options.filter_parameter_defaults
                    and (value in options.filter_parameter_defaults)
                    and (key in kwargs)
                    and (kwargs[key] == value)
                ):
                    # If the keyword argument value is one of the filtered
                    # default values, and equals the parameter default—drop it
                    # from the `kwargs` dictionary to allow the argument
                    # to be superseded by any applicable conditional defaults.
                    del kwargs[key]
                kwargs_or_defaults.setdefault(key, value)
            condition_args, condition_kwargs = (
                get_function_signature_applicable_args_kwargs(
                    condition_signature, args, kwargs_or_defaults
                )
            )
            if condition(*condition_args, **condition_kwargs):
                len_args: int = len(args)
                if len(default_args) > len_args:
                    # Extend args if there are more default args than args
                    args = (*args, *default_args[len_args:])
                kwargs.update(default_kwargs)
            return args, kwargs

        if iscoroutinefunction(function):

            @wraps(function)
            async def wrapper(*args: Any, **kwargs: Any) -> Any:
                """
                This function wraps the original and applies conditional
                defaults
                """
                args, kwargs = get_args_kwargs(*args, **kwargs)
                # Execute the wrapped function
                return await function(*args, **kwargs)

        else:

            @wraps(function)
            def wrapper(*args: Any, **kwargs: Any) -> Any:
                """
                This function wraps the original and applies conditional
                defaults
                """
                args, kwargs = get_args_kwargs(*args, **kwargs)
                # Execute the wrapped function
                return function(*args, **kwargs)

        return wrapper

    return decorating_function


@dataclass(frozen=True)
class ApplyConditionalDefaultsOptions:
    """
    This class contains options governing the behavior of the
    [apply_conditional_defaults
    ](./#decorative_secrets.defaults.apply_conditional_defaults) decorator.

    Attributes:
        filter_parameter_defaults: This may either be:

            1.  A tuple of values which should be filtered out of keyword
                arguments when both the passed keyword argument value *and*
                the parameter default value in the function signature
                are one of these values.
            2.  A dictionary mapping parameter names to tuples of values
                applying the filtering logic described above on a per-parameter
                basis.

            Defaults to `()`—indicating no filtering should occur.

    Examples:
        ```python
        from decorative_secrets.defaults import (
            apply_conditional_defaults,
            ApplyConditionalDefaultsOptions,
        )
        @apply_conditional_defaults(
            lambda environment: environment == "prod",
            ApplyConditionalDefaultsOptions(filter_parameter_defaults=(None,)),
            source_directory="/in/prod",
            target_directory="/out/prod",
        )
        @apply_conditional_defaults(
            lambda environment: environment == "dev",
            ApplyConditionalDefaultsOptions(filter_parameter_defaults=(None,)),
            source_directory="/in/dev",
            target_directory="/out/dev",
        )
        @apply_conditional_defaults(
            lambda environment: environment == "stage",
            ApplyConditionalDefaultsOptions(filter_parameter_defaults=(None,)),
            source_directory="/in/stage",
            target_directory="/out/stage",
        )
        def get_environment_source_target(
            environment: str = "dev",
            source_directory: str | None = None,
            target_directory: str | None = None,
        ) -> tuple[str, str | None, str | None]:
            return (environment, source_directory, target_directory)

        get_environment_source_target("stage", None, None)
        # ('stage', '/in/stage', '/out/stage')

        get_environment_source_target(environment="prod", None, None)
        # ('prod', '/in/prod', '/out/prod')

        get_environment_source_target(None, None, None)
        # ('dev', '/in/dev', '/out/dev')
        ```
    """

    filter_parameter_defaults: (
        tuple[Any, ...] | dict[str, tuple[Any, ...]]
    ) = ()


def _get_args_options(
    *args: Any,
) -> tuple[tuple[Any, ...], ApplyConditionalDefaultsOptions]:
    """
    This function extracts an `ApplyConditionalDefaultsOptions` instance
    from the provided arguments, if one is present.
    """
    index: int
    value: Any
    for index, value in enumerate(args):
        if isinstance(value, ApplyConditionalDefaultsOptions):
            return (*args[:index], *args[index + 1 :]), value
    return args, ApplyConditionalDefaultsOptions()



import sys
from importlib import import_module
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from types import ModuleType

from decorative_secrets import __name__ as _module_name
from decorative_secrets.utilities import get_exception_text


def _print_help() -> None:
    print(  # noqa: T201
        "Usage:\n"
        "  decorative-secrets <secret-manager> <command> [options]\n\n"
        "Secret Managers:\n"
        "  databricks\n"
        "  onepassword"
    )


def _get_command() -> str:
    command: str = ""
    if len(sys.argv) > 1:
        command = sys.argv.pop(1).lower().replace("-", "_")
    return command


def main() -> None:
    """
    Run a sub-module `main` function.
    """
    command = _get_command()
    if command in ("__help", "_h"):
        _print_help()
        return
    module: ModuleType
    try:
        try:
            module = import_module(f"{_module_name}.{command}.__main__")
        except ImportError:
            module = import_module(f"{_module_name}.{command}")
        module.main()  # type: ignore
    except ImportError:
        print(get_exception_text())  # noqa: T201
        _print_help()


if __name__ == "__main__":
    main()



from __future__ import annotations

import asyncio
import os
from dataclasses import dataclass, field
from functools import partial
from typing import TYPE_CHECKING, Any

from decorative_secrets.callback import apply_callback_arguments

if TYPE_CHECKING:
    from collections.abc import Callable, Mapping


async def _async_getenv(env: Mapping[str, str], name: str) -> str | None:
    return await asyncio.to_thread(env.__getitem__, name)


def _getenv(env: Mapping[str, str], name: str) -> str | None:
    return env[name]


@dataclass(frozen=True)
class ApplyEnvironmentArgumentsOptions:
    """
    This class contains options governing the behavior of the
    [apply_environment_arguments
    ](./#decorative_secrets.environment.apply_environment_arguments) decorator.

    Attributes:
        env: If provided, this dictionary of environment variables will be
            used in lieu of `os.environ` when retrieving environment variable
            values.
    """

    env: Mapping[str, str] = field(default_factory=lambda: os.environ)


def _get_args_options(
    *args: Any,
) -> tuple[tuple[Any, ...], ApplyEnvironmentArgumentsOptions]:
    """
    This function extracts an `ApplyEnvironmentArgumentsOptions` instance
    from the provided arguments, if one is present.
    """
    index: int
    value: Any
    for index, value in enumerate(args):
        if isinstance(value, ApplyEnvironmentArgumentsOptions):
            return (*args[:index], *args[index + 1 :]), value
    return args, ApplyEnvironmentArgumentsOptions()


def apply_environment_arguments(
    *args: ApplyEnvironmentArgumentsOptions,
    **kwargs: str,
) -> Callable:
    """
    This decorator maps parameter names to environment variables.
    Each key represents the name of a parameter in the decorated function
    which accepts an explicit input, and the corresponding mapped value is a
    parameter name accepting an environment variable from which to obtain
    the value when no value is explicitly provided.

    Parameters:
        *args: An optional [ApplyEnvironmentArgumentsOptions
            ](./#decorative_secrets.environment.ApplyEnvironmentArgumentsOptions)
            instance governing the behavior of this decorator. If not provided,
            a default instance of [ApplyEnvironmentArgumentsOptions()
            ](./#decorative_secrets.environment.ApplyEnvironmentArgumentsOptions).
            If multiple instances are provided, only the first will be used.
        **kwargs:
            A mapping of static parameter names to the parameter names
            of arguments accepting environment variable names from which to
            retrieve the value when the key argument is not explicitly
            provided.

    Example:
        ```python
        from functools import (
            cache,
        )
        from decorative_secrets.environment import (
            apply_environment_arguments,
        )
        from my_client_sdk import (
            Client,
        )


        @cache
        @apply_onepassword_arguments(
            client_id="client_id_environment_variable",
            client_secret="client_secret_environment_variable",
        )
        def get_client(
            client_id: str | None = None,
            client_secret: str = None,
            client_id_environment_variable: str | None = None,
            client_secret_environment_variable: str | None = None,
        ) -> Client:
            return Client(
                oauth2_client_id=client_id,
                oauth2_client_secret=client_secret,
            )


        client: Client = get_client(
            client_id_environment_variable=("CLIENT_ID",),
            client_secret_environment_variable=("CLIENT_SECRET",),
        )
        ```
    """
    options: ApplyEnvironmentArgumentsOptions
    _, options = _get_args_options(*args)
    return apply_callback_arguments(
        partial(_getenv, options.env),
        partial(_async_getenv, options.env),
        **kwargs,
    )



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterable


class InterfaceNotInstalledError(RuntimeError):
    """
    Raised when a required CLI is not installed, and cannot be installed
    automatically.
    """

    def __init__(self, name: str, url: str | None = None) -> None:
        message: str = (
            f"Please install [{name}]({url})"
            if url
            else f"Please install {name}"
        )
        super().__init__(message)


class OnePasswordCommandLineInterfaceNotInstalledError(
    InterfaceNotInstalledError
):
    """
    Raised when the 1Password CLI is not installed, and cannot be installed
    automatically.
    """

    def __init__(self) -> None:
        super().__init__(
            "the 1Password CLI",
            "https://developer.1password.com/docs/cli/get-started/installation",
        )


class WinGetNotInstalledError(InterfaceNotInstalledError):
    """
    Raised when WinGet is not installed on a Windows system.
    """

    def __init__(self) -> None:
        super().__init__(
            "WinGet",
            "https://learn.microsoft.com/en-us/windows/package-manager/"
            "winget/",
        )


class HomebrewNotInstalledError(InterfaceNotInstalledError):
    """
    Raised when Homebrew is not installed on a macOS system.
    """

    def __init__(self) -> None:
        super().__init__(
            "Homebrew",
            "https://brew.sh/",
        )


class DatabricksCLINotInstalledError(InterfaceNotInstalledError):
    """
    Raised when the Databricks CLI is not installed, and cannot be installed
    automatically.
    """

    def __init__(self) -> None:
        super().__init__(
            "the Databricks CLI",
            "https://docs.databricks.com/aws/en/dev-tools/cli/install",
        )


def _iter_arguments_error_messages_lines(
    arguments_error_messages: dict[str, list[str]],
) -> Iterable[str]:
    parameter_name: str
    parameter_error_messages: list[str]
    is_first: bool = True
    for (
        parameter_name,
        parameter_error_messages,
    ) in arguments_error_messages.items():
        if not is_first:
            yield ""
        yield (
            "Errors were encountered looking up values for "
            f"`{parameter_name}`:\n"
        )
        yield from parameter_error_messages
        is_first = False


class ArgumentsResolutionError(ValueError):
    """
    Raised when one or more arguments cannot be resolved.
    """

    def __init__(self, arguments_error_messages: dict[str, list[str]]) -> None:
        super().__init__(
            "\n".join(
                _iter_arguments_error_messages_lines(arguments_error_messages)
            )
        )



import asyncio
from collections import deque
from collections.abc import Callable, Coroutine, Mapping, Sequence
from functools import wraps
from inspect import Parameter, Signature, signature
from typing import Any

from decorative_secrets._utilities import (
    _FUNCTIONS_ERRORS,
    asyncio_run,
    get_errors,
    merge_function_signature_args_kwargs,
    unwrap_function,
)
from decorative_secrets.errors import ArgumentsResolutionError
from decorative_secrets.utilities import (
    get_exception_text,
    iscoroutinefunction,
)


def _get_sync_async_callbacks(
    *callbacks: Callable[..., Any],
) -> tuple[Callable[..., Any], Callable[..., Any]]:
    """
    This function validates and consolidates a set of callback functions.
    """
    message: str
    if not callbacks:
        message = "At least one callback function must be provided."
        raise ValueError(message)
    sync_callback: Callable[..., Any] | None = None
    async_callback: Callable[..., Any] | None = None
    for callback in callbacks:
        if iscoroutinefunction(callback):
            async_callback = callback
        elif callable(callback):
            sync_callback = callback
    if sync_callback is None:
        if async_callback is None:
            message = (
                "Either a `callback` or an `async_callback` argument must be "
                "provided."
            )
            raise ValueError(message)

        def sync_callback(argument: Any) -> Any:
            return asyncio_run(async_callback(argument))

    if async_callback is None:

        async def async_callback(argument: Any) -> Any:
            await asyncio.sleep(0)
            return sync_callback(argument)

    return sync_callback, async_callback


def apply_callback_arguments(  # noqa: C901
    *callbacks: Callable[..., Any],
    **callback_parameter_names: str,
) -> Callable[..., Callable[..., Any]]:
    """
    This decorator maps parameter names to callback arguments.
    Each key represents the name of a parameter in the decorated function
    which accepts an explicit input, and the corresponding mapped value is
    an argument to pass to the provided callback function(s).

    Parameters:
        *callbacks: One or more callback functions. If both synchronous and
            asynchronous functions are provided, they will be used
            appropriately based on the decorated function's type, otherwise
            synchronous functions will be wrapped for asynchronous use
            and vice versa.
        **callback_parameter_names:
            A mapping of static parameter names to callback parameter names.

    Returns:
        A decorator function which retrieves argument values by
            passing callback function arguments to the callback, and
            applying the output to their mapped static parameters.

    Examples:
        >>> @apply_callback_arguments(
        ...     lambda x: x * 2,
        ...     {"x": "x_lookup_args"},
        ... )
        ... def return_value(
        ...     x: int | None = None,
        ...     x_lookup_args: tuple[
        ...         Sequence[int],
        ...         Mapping[str, int],
        ...     ]
        ...     | None = None,
        ... ) -> int:
        ...     return x**2
        >>> return_value(
        ...     x_lookup_args=(
        ...         3,
        ...         None,
        ...     )
        ... )
        36
    """
    callback: Callable[..., Any]
    async_callback: Callable[..., Any]
    callback, async_callback = _get_sync_async_callbacks(*callbacks)

    def decorating_function(  # noqa: C901
        function: Callable[..., Any],
    ) -> Callable[..., Any]:
        original_function: Callable[..., Any] = unwrap_function(function)
        function_signature: Signature = signature(original_function)

        def get_args_kwargs(  # noqa: C901
            *args: Any, **kwargs: Any
        ) -> tuple[tuple[Any, ...], dict[str, Any]]:
            """
            This function performs lookups for any parameters for which an
            argument is not passed explicitly.
            """
            # Capture errors
            errors: dict[str, list[str]] = get_errors(original_function)
            # First we consolidate the keyword arguments with any arguments
            # which are passed to parameters which can be either positional
            # *or* keyword arguments, and were passed as positional arguments
            args = merge_function_signature_args_kwargs(
                function_signature, args, kwargs
            )
            # For any arguments where we have callback arguments and do not
            # have an explicitly passed value, execute the callback
            key: str
            value: Any
            used_keys: set[str] = {
                key for key, value in kwargs.items() if value is not None
            }
            unused_callback_parameter_names: set[str] = (
                set(callback_parameter_names.values()) & used_keys
            )
            parameter_name: str
            for parameter_name in (
                set(callback_parameter_names.keys()) - used_keys
            ):
                callback_parameter_name: str = callback_parameter_names[
                    parameter_name
                ]
                unused_callback_parameter_names.discard(
                    callback_parameter_name
                )
                callback_argument: Any = kwargs.pop(
                    callback_parameter_name, None
                )
                parameter: Parameter | None = (
                    function_signature.parameters.get(parameter_name)
                )
                callback_: Callable[..., Any] = callback
                if (
                    (parameter is not None)
                    and (isinstance(parameter.annotation, type))
                    and issubclass(Coroutine, parameter.annotation)
                ):
                    callback_ = async_callback
                if callback_argument is not None:
                    try:
                        kwargs[parameter_name] = callback_(callback_argument)
                        # Clear preceding errors for this parameter
                        errors.pop(parameter_name, None)
                    except Exception:  # noqa: BLE001
                        errors.setdefault(parameter_name, [])
                        errors[parameter_name].append(get_exception_text())
                elif callback_parameter_name in function_signature.parameters:
                    default: tuple[Sequence[Any], Mapping[str, Any]] | None = (
                        function_signature.parameters[
                            callback_parameter_name
                        ].default
                    )
                    if default not in (Signature.empty, None):
                        try:
                            kwargs[parameter_name] = callback_(default)
                            # Clear preceding errors for this parameter
                            errors.pop(parameter_name, None)
                        except Exception:  # noqa: BLE001
                            errors.setdefault(parameter_name, [])
                            errors[parameter_name].append(get_exception_text())
                if (function is original_function) and errors:
                    arguments_error_messages: dict[str, list[str]] = {}
                    for key, argument_error_messages in errors.items():
                        # Don't raise an error for parameters which
                        # have a value or default value
                        if kwargs.get(key) is None:
                            parameter = function_signature.parameters.get(key)
                            if parameter and (
                                parameter.default is Signature.empty
                            ):
                                arguments_error_messages[key] = (
                                    argument_error_messages
                                )
                    # Clear global errors collection
                    _FUNCTIONS_ERRORS.pop(id(function), None)
                    if arguments_error_messages:
                        raise ArgumentsResolutionError(
                            arguments_error_messages
                        )
            # Remove unused callback arguments
            deque(map(kwargs.pop, unused_callback_parameter_names), maxlen=0)
            return (args, kwargs)

        if iscoroutinefunction(function):

            @wraps(function)
            async def wrapper(*args: Any, **kwargs: Any) -> Any:
                """
                This function wraps the original and performs lookups for
                any parameters for which an argument is not passed
                """
                args, kwargs = get_args_kwargs(*args, **kwargs)
                # Execute the wrapped function
                return await function(*args, **kwargs)

        else:

            @wraps(function)
            def wrapper(*args: Any, **kwargs: Any) -> Any:
                """
                This function wraps the original and performs lookups for
                any parameters for which an argument is not passed
                """
                args, kwargs = get_args_kwargs(*args, **kwargs)
                # Execute the wrapped function
                return function(*args, **kwargs)

        return wrapper

    return decorating_function



from __future__ import annotations

import asyncio
import os
import sys
from contextlib import suppress
from functools import cache, update_wrapper
from inspect import Parameter, Signature, signature
from io import TextIOWrapper
from shutil import which
from subprocess import (
    CalledProcessError,
)
from typing import TYPE_CHECKING, Any
from urllib.request import urlopen

import nest_asyncio  # type: ignore[import-untyped]

from decorative_secrets.errors import (
    HomebrewNotInstalledError,
    WinGetNotInstalledError,
)
from decorative_secrets.subprocess import check_output
from decorative_secrets.utilities import as_dict

if TYPE_CHECKING:
    from collections.abc import Callable, Coroutine, Iterable, Sequence


HOMEBREW_INSTALL_SH: str = (
    "https://raw.githubusercontent.com/Homebrew/install/HEAD/install.sh"
)


def install_brew() -> None:
    """
    Install Homebrew on macOS or linux if not already installed.
    """
    env: dict[str, str] = os.environ.copy()
    env["NONINTERACTIVE"] = "1"
    bash: str = which("bash") or "/bin/bash"
    with TextIOWrapper(
        urlopen(HOMEBREW_INSTALL_SH)  # noqa: S310
    ) as response_io:
        try:
            check_output(
                (bash, "-c", response_io.read()),
                env=env,
            )
        except CalledProcessError as error:
            # This is usually because the script requires `sudo` access to run
            raise HomebrewNotInstalledError from error


@cache
def which_brew() -> str:
    """
    Find the `brew` executable on macOS, or install Homebrew if not found.
    """
    brew: str | None
    brew = which("brew") or "brew"
    try:
        check_output((brew, "--version"))
    except (CalledProcessError, FileNotFoundError):
        install_brew()
        brew = which("brew")
        if not brew:
            if sys.platform == "darwin":
                brew = "/opt/homebrew/bin/brew"
                if not os.path.exists(brew):
                    brew = "brew"
            else:
                brew = "/home/linuxbrew/.linuxbrew/bin/brew"
                if not os.path.exists(brew):
                    brew = "brew"
        try:
            check_output((brew, "--version"))
        except (CalledProcessError, FileNotFoundError) as error:
            raise HomebrewNotInstalledError from error
    return brew


@cache
def which_winget() -> str | None:
    """
    Find the `winget` executable on Windows, or raise an error if not found.
    """
    winget: str = which("winget") or "winget"
    try:
        check_output((winget, "--version"))
    except (CalledProcessError, FileNotFoundError) as error:
        raise WinGetNotInstalledError from error
    else:
        return winget


def as_tuple(
    user_function: Callable[..., Iterable[Any]],
) -> Callable[..., tuple[Any, ...]]:
    """
    This is a decorator which will return an iterable as a tuple.
    """

    def wrapper(*args: Any, **kwargs: Any) -> tuple[Any, ...]:
        return tuple(user_function(*args, **kwargs) or ())

    return update_wrapper(wrapper, user_function)


@as_tuple
def merge_function_signature_args_kwargs(
    function_signature: Signature, args: Iterable[Any], kwargs: dict[str, Any]
) -> Iterable[Any]:
    """
    This function merges positional/keyword arguments for a function
    into the keyword argument dictionary, and returns any arguments which
    are positional-only.
    """
    value: Any
    parameter: Parameter
    if args:
        for parameter, value in zip(
            function_signature.parameters.values(), args, strict=False
        ):
            if parameter.kind == Parameter.POSITIONAL_OR_KEYWORD:
                kwargs[parameter.name] = value
            else:
                yield value


@as_dict
def get_signature_parameter_names_defaults(
    function_signature: Signature,
) -> Iterable[tuple[str, Any]]:
    """
    This function returns a dictionary mapping parameter names to their default
    values for all keyword parameters in the function signature.
    """
    parameter: Parameter
    for parameter in function_signature.parameters.values():
        if (parameter.default is not Signature.empty) and parameter.name:
            yield parameter.name, parameter.default


def get_function_signature_applicable_args_kwargs(
    function_signature: Signature | Callable,
    args: Sequence[Any],
    kwargs: dict[str, Any],
) -> tuple[tuple[Any, ...], dict[str, Any]]:
    """
    Given a function or function signature, and positional and keyword
    arguments, this function returns only those arguments and keyword
    arguments which are applicable to the function.

    Parameters:
        function_signature: A function or function signature whose
            parameters will be used to filter the provided arguments.
        args: A sequence of positional arguments.
        kwargs: A dictionary of keyword arguments.
    """
    applicable_kwargs: dict[str, Any] = {}
    max_positional_argument_count: int | None = 0
    parameter: Parameter
    if not isinstance(function_signature, Signature):
        function_signature = signature(function_signature)
    for parameter in function_signature.parameters.values():
        if parameter.kind == Parameter.VAR_KEYWORD:
            # All keywords are accepted
            applicable_kwargs = kwargs
            break
        if parameter.kind in (
            Parameter.KEYWORD_ONLY,
            Parameter.POSITIONAL_OR_KEYWORD,
        ) and (parameter.name in kwargs):
            applicable_kwargs[parameter.name] = kwargs[parameter.name]
        elif max_positional_argument_count is not None:
            if parameter.kind in (
                Parameter.POSITIONAL_ONLY,
                Parameter.POSITIONAL_OR_KEYWORD,
            ):
                max_positional_argument_count += 1
            elif parameter.kind == Parameter.VAR_POSITIONAL:
                # Unlimited positional arguments
                max_positional_argument_count = None
    return (tuple(args[:max_positional_argument_count]), applicable_kwargs)


def get_running_loop() -> asyncio.AbstractEventLoop | None:
    """
    Get the currently running event loop, or None if there is none.
    """
    loop: asyncio.AbstractEventLoop | None = None
    with suppress(RuntimeError):
        loop = asyncio.get_running_loop()
    return loop


def asyncio_run(coroutine: Coroutine) -> Any:
    """
    Run a coroutine, applying nest_asyncio if necessary.
    """
    loop: asyncio.AbstractEventLoop | None = get_running_loop()
    if loop is None:
        return asyncio.run(coroutine)
    nest_asyncio.apply(loop)
    return asyncio.run(coroutine)


def unwrap_function(
    function: Callable[..., Any],
) -> Callable:
    """
    This function retrieves the original, unwrapped, decorated function.
    """
    while hasattr(function, "__wrapped__"):
        function = function.__wrapped__
    return function


_FUNCTIONS_ERRORS: dict[int, dict[str, list[str]]] = {}


def get_errors(function: Callable[..., Any]) -> dict[str, list[str]]:
    """
    This function retrieves the current function errors.
    """
    function_id: int = id(function)
    return _FUNCTIONS_ERRORS.setdefault(function_id, {})



import asyncio
import inspect
import logging
import sys
from collections.abc import Awaitable, Callable, Iterable, Iterator
from functools import partial, wraps
from time import sleep
from traceback import format_exception
from typing import Any, Protocol, overload
from warnings import warn


def iscoroutinefunction(function: Any) -> bool:
    """
    An adaptation of `asyncio.iscoroutinefunction`
    """
    if isinstance(function, partial):
        return iscoroutinefunction(function.func)
    return (
        inspect.iscoroutinefunction(function)
        or type(getattr(function, "_is_coroutine", None)) is object
    )


def as_tuple(
    function: Callable[..., Iterable[Any] | Awaitable[Iterable[Any]]],
) -> Callable[..., Any]:
    """
    This is a decorator which will return an iterable as a tuple.

    Examples:
        ```python
        from decorative_secrets.utilities import as_tuple


        @as_tuple
        def get_numbers() -> Iterable[int]:
            yield 1
            yield 2
            yield 3


        assert get_numbers() == (1, 2, 3)
        ```
    """
    if iscoroutinefunction(function):

        @wraps(function)
        async def wrapper(*args: Any, **kwargs: Any) -> tuple[Any, ...]:
            return tuple(  # --
                await function(*args, **kwargs) or ()  # type: ignore[misc]
            )

    else:

        @wraps(function)
        def wrapper(*args: Any, **kwargs: Any) -> tuple[Any, ...]:
            return tuple(
                function(*args, **kwargs) or ()  # type: ignore[arg-type]
            )

    return wrapper


@overload
def as_str(
    function: None = None,
    separator: str = "",
) -> Callable[..., Callable[..., str]]: ...


@overload
def as_str(
    function: Callable[..., Iterable[str]] = ...,
    separator: str = "",
) -> Callable[..., str]: ...


def as_str(
    function: Callable[..., Iterable[str]]
    | Awaitable[Iterable[Any]]
    | None = None,
    separator: str = "",
) -> Callable[..., Callable[..., str]] | Callable[..., str]:
    """
    This decorator causes a function yielding an iterable of strings to
    return a single string with the elements joined by the specified
    `separator`.

    Parameters:
        function: The function to decorate. If `None`, a decorating
            function is returned.
        separator: The string used to join the iterable elements.

    Returns:
        A decorator which joins the iterable elements into a single string.

    Examples:
        ```python
        from decorative_secrets.utilities import as_str


        @as_str(separator=", ")
        def get_fruits() -> Iterable[str]:
            yield "apple"
            yield "banana"
            yield "cherry"


        assert get_fruits() == "apple, banana, cherry"
        ```

        ```python
        from decorative_secrets.utilities import as_str


        @as_str
        def get_fruits() -> Iterable[str]:
            yield "apple\n"
            yield "banana\n"
            yield "cherry"


        assert get_fruits() == "apple\nbanana\ncherry"
        ```
    """

    def decorating_function(
        user_function: Callable[..., Iterable[str]],
    ) -> Callable[..., Any]:
        if iscoroutinefunction(user_function):

            @wraps(user_function)
            async def wrapper(*args: Any, **kwargs: Any) -> str:
                return separator.join(
                    await user_function(  # type: ignore[misc]
                        *args, **kwargs
                    )
                    or ()
                )

        else:

            @wraps(user_function)
            def wrapper(*args: Any, **kwargs: Any) -> str:
                return separator.join(user_function(*args, **kwargs) or ())

        return wrapper

    if function is None:
        return decorating_function
    return decorating_function(function)  # type: ignore[arg-type]


def as_dict(
    function: Callable[
        ..., Iterable[tuple[Any, Any]] | Awaitable[Iterable[tuple[Any, Any]]]
    ],
) -> Callable[..., Any]:
    """
    This is a decorator which will return an iterable of key/value pairs
    as a dictionary.

    Examples:
        ```python
        from decorative_secrets.utilities import as_dict


        @as_dict
        def get_settings() -> Iterable[tuple[str, Any]]:
            yield ("host", "localhost")
            yield ("port", 8080)
            yield ("debug", True)


        assert get_settings() == (
            {"host": "localhost", "port": 8080, "debug": True}
        )
        ```
    """

    if iscoroutinefunction(function):

        @wraps(function)
        async def wrapper(*args: Any, **kwargs: Any) -> dict[Any, Any]:
            return dict(
                await function(*args, **kwargs) or ()  # type: ignore[misc]
            )

    else:

        @wraps(function)
        def wrapper(*args: Any, **kwargs: Any) -> dict[Any, Any]:
            return dict(
                function(*args, **kwargs) or ()  # type: ignore[arg-type]
            )

    return wrapper


def as_iter(
    function: Callable[..., Iterable[Any]],
) -> Callable[..., Any]:
    """
    This is a decorator which will return an iterator for a function
    yielding an iterable.

    Examples:
        ```python
        from decorative_secrets.utilities import as_iter
        from collections.abc import Iterator


        @as_iter
        def get_settings() -> Iterable[tuple[str, Any]]:
            yield ("host", "localhost")
            yield ("port", 8080)
            yield ("debug", True)


        assert issubclass(get_settings(), Iterator)
        ```
    """

    if iscoroutinefunction(function):

        @wraps(function)
        async def wrapper(*args: Any, **kwargs: Any) -> Iterator[Any]:
            return iter(
                await function(*args, **kwargs) or ()  # type: ignore[misc]
            )

    else:

        @wraps(function)
        def wrapper(*args: Any, **kwargs: Any) -> Iterator[Any]:
            return iter(function(*args, **kwargs) or ())

    return wrapper


def _default_retry_hook(
    error: Exception,
    attempt_number: int,  # noqa: ARG001
    *args: Any,  # noqa: ARG001
    **kwargs: Any,  # noqa: ARG001
) -> bool:
    """
    This is a retry hook which simply returns `True` for any error
    instance, allowing all retries to proceed.
    """
    if not error:
        raise ValueError(error)
    return True


class RetryHook(Protocol):
    def __call__(
        self, error: Exception, *args: Any, **kwargs: Any
    ) -> bool: ...


class AsyncRetryHook(Protocol):
    async def __call__(
        self, error: Exception, *args: Any, **kwargs: Any
    ) -> bool: ...


def warn_retry_hook(
    error: Exception,
    attempt_number: int,
    *args: Any,  # noqa: ARG001
    **kwargs: Any,  # noqa: ARG001
) -> bool:
    """
    This is a retry hook which will issue a warning and retry number
    whenever an error occurs.
    """
    message: str = f"Attempt # {attempt_number} failed with error: {error}"
    warn(
        message,
        stacklevel=2,
    )
    return True


def create_log_warning_retry_hook(
    logger: logging.Logger | Callable[[], logging.Logger],
) -> RetryHook:
    """
    This factory creates a retry hook which logs warning using the provided
    logger.

    Parameters:
        logger: The logger to use for logging warnings, or a callable which
        returns a logger.
    """
    if not isinstance(logger, logging.Logger) and callable(logger):
        logger = logger()
    if not isinstance(logger, logging.Logger):
        raise TypeError(logger)

    def retry_hook(
        error: Exception,
        attempt_number: int,
        *args: Any,  # noqa: ARG001
        **kwargs: Any,  # noqa: ARG001
    ) -> bool:
        logger.warning(
            "Attempt # %d failed with error: %s",
            attempt_number,
            str(error),
            stacklevel=2,
        )
        return True

    return retry_hook


def create_async_log_warning_retry_hook(
    logger: logging.Logger,
) -> AsyncRetryHook:
    """
    This factory creates an async retry hook which logs warning using the
    provided logger.

    !!! Note
        Please make sure to use a [non-blocking logger
        ](https://docs.python.org/3/howto/logging-cookbook.html#dealing-with-handlers-that-block).

    Parameters:
        logger: The logger to use for logging warnings.
    """

    async def retry_hook(
        error: Exception,
        attempt_number: int,
        *args: Any,  # noqa: ARG001
        **kwargs: Any,  # noqa: ARG001
    ) -> bool:
        logger.warning(
            "Attempt # %d failed with error: %s",
            attempt_number,
            str(error),
            stacklevel=2,
        )
        return True

    return retry_hook


def retry(  # noqa: C901
    errors: tuple[type[Exception], ...],
    retry_hook: RetryHook | AsyncRetryHook = _default_retry_hook,
    number_of_attempts: int = 2,
) -> Callable:
    """
    This is a decorator which will retry a function a specified
    number of times, with exponential backoff, if it raises one of the
    specified errors types.

    Parameters:
        errors: A tuple of exception types which should trigger a retry.
        retry_hook: A function which is called with the exception instance
            (optionally) and an attempt number when an error occurs. If this
            function returns `False`, the exception is re-raised and no further
            retries are attempted.
        number_of_attempts: The total number of attempts to make, including
            the initial attempt.
    """

    def decorating_function(function: Callable) -> Callable:
        attempt_number: int = 1
        if iscoroutinefunction(function):

            @wraps(function)
            async def wrapper(*args: Any, **kwargs: Any) -> Any:
                nonlocal attempt_number
                if (number_of_attempts - attempt_number) > 0:
                    # If `number_of_attempts` is greater than `attempt_number`,
                    # we have remaining attempts to try, so catch errors.
                    try:
                        return await function(*args, **kwargs)
                    except errors as error:
                        if not (
                            (
                                await retry_hook(  # type: ignore[misc]
                                    error, attempt_number
                                )
                                if len(
                                    inspect.signature(retry_hook).parameters
                                )
                                > 1
                                else await retry_hook(  # type: ignore[misc]
                                    error
                                )
                            )
                            if iscoroutinefunction(retry_hook)
                            else (
                                retry_hook(error, attempt_number)
                                if len(
                                    inspect.signature(retry_hook).parameters
                                )
                                > 1
                                else retry_hook(error)
                            )
                        ):
                            raise
                        await asyncio.sleep(2**attempt_number)
                        attempt_number += 1
                        return await wrapper(*args, **kwargs)
                # This is our last attempt, so just call the function.
                return await function(*args, **kwargs)

        else:

            @wraps(function)
            def wrapper(*args: Any, **kwargs: Any) -> Any:
                nonlocal attempt_number
                if (number_of_attempts - attempt_number) > 0:
                    try:
                        return function(*args, **kwargs)
                    except errors as error:
                        if not (
                            retry_hook(error, attempt_number)
                            if len(inspect.signature(retry_hook).parameters)
                            > 1
                            else retry_hook(error)
                        ):
                            raise
                        sleep(2**attempt_number)
                        attempt_number += 1
                        return wrapper(*args, **kwargs)
                return function(*args, **kwargs)

        return wrapper

    return decorating_function


def get_exception_text() -> str:
    """
    When called within an exception, this function returns a text
    representation of the error matching what is found in
    `traceback.print_exception`, but is returned as a string value rather than
    printing.
    """
    return "".join(format_exception(*sys.exc_info()))



from __future__ import annotations

import argparse
import copyreg
import inspect
import json
import os
import sys
from contextlib import suppress
from dataclasses import asdict, dataclass
from functools import cache, partial
from shutil import which
from subprocess import CalledProcessError
from typing import TYPE_CHECKING, Any, TypedDict
from urllib.request import urlopen

from databricks.sdk import WorkspaceClient

from decorative_secrets._utilities import (
    which_brew,
    which_winget,
)
from decorative_secrets.callback import apply_callback_arguments
from decorative_secrets.errors import (
    DatabricksCLINotInstalledError,
    HomebrewNotInstalledError,
)
from decorative_secrets.subprocess import check_call, check_output
from decorative_secrets.utilities import retry

if TYPE_CHECKING:
    from collections.abc import Callable, Mapping

    from databricks.sdk.config import Config
    from databricks.sdk.credentials_provider import CredentialsStrategy
    from databricks.sdk.dbutils import RemoteDbUtils
    from databricks.sdk.oauth import AuthorizationDetail


# region Make workspace clients pickleable (in most scenarios)


def _unpickle_workspace_client(
    host: str,
    client_id: str | None,
    client_secret: str | None,
    token: str | None,
) -> WorkspaceClient:
    if client_id and client_secret:
        return WorkspaceClient(
            host=host,
            client_id=client_id,
            client_secret=client_secret,
        )
    return WorkspaceClient(
        host=host,
        token=token,
    )


def _workspace_client_redux(
    client: WorkspaceClient,
) -> tuple[Callable[..., WorkspaceClient], tuple]:
    token: str | None = None
    with suppress(ValueError, AttributeError):
        token = client.config.oauth_token().access_token
    return (
        _unpickle_workspace_client,
        (
            client.config.host,
            client.config.client_id,
            client.config.client_secret,
            token,
        ),
    )


copyreg.pickle(
    WorkspaceClient,
    _workspace_client_redux,
)

# endregion


@dataclass
class DatabricksWorkspaceClientArguments:
    """
    An object holding arguments to pass to a Databricks workspace client
    if/when retrieving secrets remotely.
    """

    host: str | None = None
    account_id: str | None = None
    username: str | None = None
    password: str | None = None
    client_id: str | None = None
    client_secret: str | None = None
    token: str | None = None
    profile: str | None = None
    config_file: str | None = None
    azure_workspace_resource_id: str | None = None
    azure_client_secret: str | None = None
    azure_client_id: str | None = None
    azure_tenant_id: str | None = None
    azure_environment: str | None = None
    auth_type: str | None = None
    cluster_id: str | None = None
    google_credentials: str | None = None
    google_service_account: str | None = None
    debug_truncate_bytes: int | None = None
    debug_headers: bool | None = None
    product: str | None = None
    product_version: str | None = None
    credentials_strategy: CredentialsStrategy | None = None
    credentials_provider: CredentialsStrategy | None = None
    token_audience: str | None = None


def apply_databricks_secrets_arguments(
    *args: DatabricksWorkspaceClientArguments,
    **kwargs: str,
) -> Callable:
    """
    This decorator maps parameter names to Databricks secrets.
    Each key in `databricks_secret_arguments` represents the name of a
    parameter in the decorated function which accepts an explicit input, and
    the corresponding mapped value is a parameter name accepting a tuple with
    the secret scope and key with which to lookup a secret to pass to the
    mapped parameter in lieu of an explicitly provided argument.

    Parameters:
        *args: A `DatabricksWorkspaceConfigArguments` instance to configure
            a workspace client when retrieving secrets remotely
            (if more than one  is provided, only the first is used).
        **kwargs: A mapping of static parameter names to the parameter names
            of arguments accepting Databricks secret scope + key tuples
            from which to retrieve a value when the key argument is not
            explicitly provided.

    Example:
        ```python
        from functools import (
            cache,
        )
        from decorative_secrets.databricks import (
            apply_databricks_secret_arguments,
        )
        from my_client_sdk import (
            Client,
        )


        @cache
        @apply_databricks_secret_arguments(
            client_id="client_id_databricks_secret",
            client_secret="client_secret_databricks_secret",
        )
        def get_client(
            client_id: str | None = None,
            client_secret: str = None,
            client_id_databricks_secret: str | None = None,
            client_secret_databricks_secret: str | None = None,
        ) -> Client:
            return Client(
                oauth2_client_id=client_id,
                oauth2_client_secret=client_secret,
            )


        client: Client = get_client(
            client_id_databricks_secret=(
                "client",
                "client-id",
            ),
            client_secret_databricks_secret=(
                "client",
                "client-secret",
            ),
        )
        ```
    """
    databricks_workspace_client_arguments: (
        DatabricksWorkspaceClientArguments | None
    ) = _get_args_options(*args)[1]
    get_scope_key_secret: Callable[[str | tuple[str, str]], str] = (
        partial(
            _get_scope_key_secret,
            **asdict(databricks_workspace_client_arguments),
        )
        if databricks_workspace_client_arguments
        else _get_scope_key_secret
    )
    return apply_callback_arguments(
        get_scope_key_secret,
        **kwargs,
    )


def _install_sh_databricks_cli() -> None:
    """
    Install the Databricks CLI using the install script.
    """
    with urlopen(
        "https://raw.githubusercontent.com/databricks/setup-cli/"
        "main/install.sh"
    ) as install_io:
        sh: str = which("sh") or "sh"
        try:
            check_call((sh,), input=install_io.read())
        except (CalledProcessError, FileNotFoundError) as error:
            if (
                (not isinstance(error, CalledProcessError))
                or (not error.stdout)
                or (
                    (b"already exists" not in error.stdout)
                    and (b"'sudo'" not in error.stdout)
                )
            ):
                # This is usually because the script requires `sudo` access to
                # run
                raise DatabricksCLINotInstalledError from error


def _install_databricks_cli() -> None:
    """
    Install the Databricks CLI.
    """
    if sys.platform.startswith("win"):
        winget: str | None = which_winget()
        if winget:
            with suppress(CalledProcessError):
                check_output((winget, "search", "DatabricksCLI"))
            with suppress(CalledProcessError):
                check_output((winget, "install", "Databricks.DatabricksCLI"))
                return
    elif sys.platform == "darwin":
        brew: str
        # Here we suppress the HomebrewNotInstalledError because we
        # can still attempt to install the Databricks CLI using
        # the install script
        with suppress(HomebrewNotInstalledError):
            brew = which_brew()
            if brew:
                with suppress(CalledProcessError):
                    check_output((brew, "tap", "databricks/tap"))
                with suppress(CalledProcessError):
                    check_output((brew, "install", "databricks"))
                    return
    _install_sh_databricks_cli()


def which_databricks() -> str:
    """
    Find the `databricks` executable, or install the Databricks CLI if not
    found.
    """
    databricks: str = which("databricks") or "databricks"
    try:
        check_output((databricks, "--version"))
    except (CalledProcessError, FileNotFoundError):
        _install_databricks_cli()
        databricks = which("databricks") or "databricks"
    return databricks


class _DatabricksAuthProfile(TypedDict, total=False):
    name: str
    host: str
    cloud: str
    auth_type: str
    valid: bool


class _DatabricksAuthProfiles(TypedDict, total=True):
    profiles: list[_DatabricksAuthProfile]


class _DatabricksAuthDescription(TypedDict, total=False):
    status: str
    username: str
    details: dict[str, Any]
    error: dict[str, Any]


def _get_host_profile(
    host: str,
) -> str | None:
    host = host.lower()
    auth_profile: _DatabricksAuthProfile
    for auth_profile in _databricks_auth_profiles()["profiles"]:
        if auth_profile.get("host", "").lower() == host:
            return auth_profile.get("name")
    return None


def _databricks_auth_profiles() -> _DatabricksAuthProfiles:
    databricks: str = which_databricks()
    return json.loads(
        check_output(
            (databricks, "auth", "profiles", "-o", "json"),
            input=b"\n\n",
        )
    )


def _databricks_auth_describe(
    host: str | None = None,
    profile: str | None = None,
    target: str | None = None,
) -> _DatabricksAuthDescription:
    if host and not profile:
        profile = _get_host_profile(host)
    databricks: str = which_databricks()
    output: str
    if host or profile or target:
        output = check_output(
            (
                databricks,
                "auth",
                "describe",
                "-o",
                "json",
                *(("--host", host) if host else ()),
                *(("--profile", profile) if profile else ()),
                *(("--target", target) if target else ()),
            ),
            input=b"\n\n",
        )
    else:
        # Automatically select the default/first profile if no host,
        # profile, or target is specified
        output = check_output(
            (databricks, "auth", "describe", "-o", "json"),
            input=b"\n\n",
        )
    return json.loads(output)


@cache
@retry(
    (CalledProcessError,),
    number_of_attempts=3,
)
def _databricks_auth_login(
    host: str | None = None,
    profile: str | None = None,
    target: str | None = None,
) -> None:
    if host and not profile:
        profile = _get_host_profile(host)
    databricks: str = which_databricks()
    if host or profile or target:
        check_call(
            (
                databricks,
                "auth",
                "login",
                *(("--host", host) if host else ()),
                *(("--profile", profile) if profile else ()),
                *(("--target", target) if target else ()),
            ),
            input=b"\n\n",
        )
    else:
        # Automatically select the default/first profile if no host,
        # profile, or target is specified
        check_call((databricks, "auth", "login"), input=b"\n\n")


def databricks_auth_login(
    host: str | None = None,
    profile: str | None = None,
    target: str | None = None,
) -> None:
    """
    Log in to Databricks using the CLI if not already logged in.

    Parameters:
        host: A Databricks workspace host URL.
        profile: A Databricks Configuration Profile.
        target: A Databricks CLI target.
    """
    if (host is None) and (profile is None) and (target is None):
        host = os.getenv("DATABRICKS_HOST")
        profile = os.getenv("DATABRICKS_CONFIG_PROFILE")
    # with suppress(CalledProcessError):
    # If we are already authenticated, don't attempt to log in again
    with suppress(CalledProcessError):
        if (
            _databricks_auth_describe(
                host=host, profile=profile, target=target
            ).get("status")
            == "success"
        ):
            return
    return _databricks_auth_login(host=host, profile=profile, target=target)


@cache
def _get_env_databricks_workspace_client(
    host: str | None = None,
    account_id: str | None = None,
    username: str | None = None,
    password: str | None = None,
    client_id: str | None = None,
    client_secret: str | None = None,
    token: str | None = None,
    profile: str | None = None,
    config_file: str | None = None,
    azure_workspace_resource_id: str | None = None,
    azure_client_secret: str | None = None,
    azure_client_id: str | None = None,
    azure_tenant_id: str | None = None,
    azure_environment: str | None = None,
    auth_type: str | None = None,
    cluster_id: str | None = None,
    google_credentials: str | None = None,
    google_service_account: str | None = None,
    debug_truncate_bytes: int | None = None,
    *,
    debug_headers: bool | None = None,
    product: str | None = None,
    product_version: str | None = None,
    credentials_strategy: CredentialsStrategy | None = None,
    credentials_provider: CredentialsStrategy | None = None,
    token_audience: str | None = None,
    config: Config | None = None,
    scopes: list[str] | None = None,
    authorization_details: list[AuthorizationDetail] | None = None,
    **env: str,  # noqa: ARG001
) -> WorkspaceClient:
    """
    Get a Databricks WorkspaceClient. This function is cached based on
    environment variables, to ensure changes to the environment are reflected.
    """
    if config:
        host = host or config.host
        profile = profile or config.profile
    if not (
        token
        or (
            (
                client_id
                or os.getenv("DATABRICKS_CLIENT_ID")
                or (config and config.client_id)
            )
            and (
                client_secret
                or os.getenv("DATABRICKS_CLIENT_SECRET")
                or (config and config.client_secret)
            )
        )
    ):
        with suppress(
            CalledProcessError,
            FileNotFoundError,
            DatabricksCLINotInstalledError,
        ):
            databricks_auth_login(host=host, profile=profile)
    environ: Mapping[str, str] = os.environ.copy() if profile else os.environ
    try:
        if profile:
            # If a profile was explicitly provided, ensure it is used,
            # and environment variables are ignored
            key: str
            for key in os.environ:
                if key.startswith("DATABRICKS_"):
                    os.environ.pop(key, None)
        parameter_name: str
        argument: Any
        client: WorkspaceClient = WorkspaceClient(
            **{  # type: ignore[arg-type]
                parameter_name: argument
                for parameter_name, argument in (
                    ("host", host),
                    ("account_id", account_id),
                    ("username", username),
                    ("password", password),
                    ("client_id", client_id),
                    ("client_secret", client_secret),
                    ("token", token),
                    ("profile", profile),
                    ("config_file", config_file),
                    (
                        "azure_workspace_resource_id",
                        azure_workspace_resource_id,
                    ),
                    ("azure_client_secret", azure_client_secret),
                    ("azure_client_id", azure_client_id),
                    ("azure_tenant_id", azure_tenant_id),
                    ("azure_environment", azure_environment),
                    ("auth_type", auth_type),
                    ("cluster_id", cluster_id),
                    ("google_credentials", google_credentials),
                    ("google_service_account", google_service_account),
                    ("debug_truncate_bytes", debug_truncate_bytes),
                    ("debug_headers", debug_headers),
                    ("product", product),
                    ("product_version", product_version),
                    ("credentials_strategy", credentials_strategy),
                    ("credentials_provider", credentials_provider),
                    ("token_audience", token_audience),
                    ("config", config),
                    ("scopes", scopes),
                    ("authorization_details", authorization_details),
                )
                if parameter_name
                in inspect.signature(WorkspaceClient.__init__).parameters
                and (argument is not None)
            }
        )
    finally:
        if profile:
            # Restore the original environment if a profile was used
            os.environ.update(environ)
    return client


def get_databricks_workspace_client(
    host: str | None = None,
    account_id: str | None = None,
    username: str | None = None,
    password: str | None = None,
    client_id: str | None = None,
    client_secret: str | None = None,
    token: str | None = None,
    profile: str | None = None,
    config_file: str | None = None,
    azure_workspace_resource_id: str | None = None,
    azure_client_secret: str | None = None,
    azure_client_id: str | None = None,
    azure_tenant_id: str | None = None,
    azure_environment: str | None = None,
    auth_type: str | None = None,
    cluster_id: str | None = None,
    google_credentials: str | None = None,
    google_service_account: str | None = None,
    debug_truncate_bytes: int | None = None,
    *,
    debug_headers: bool | None = None,
    product: str | None = None,
    product_version: str | None = None,
    credentials_strategy: CredentialsStrategy | None = None,
    credentials_provider: CredentialsStrategy | None = None,
    token_audience: str | None = None,
    config: Config | None = None,
) -> WorkspaceClient:
    """
    Get a Databricks WorkspaceClient configured from environment variables.
    """
    return _get_env_databricks_workspace_client(
        host=host,
        account_id=account_id,
        username=username,
        password=password,
        client_id=client_id,
        client_secret=client_secret,
        token=token,
        profile=profile,
        config_file=config_file,
        azure_workspace_resource_id=azure_workspace_resource_id,
        azure_client_secret=azure_client_secret,
        azure_client_id=azure_client_id,
        azure_tenant_id=azure_tenant_id,
        azure_environment=azure_environment,
        auth_type=auth_type,
        cluster_id=cluster_id,
        google_credentials=google_credentials,
        google_service_account=google_service_account,
        debug_truncate_bytes=debug_truncate_bytes,
        debug_headers=debug_headers,
        product=product,
        product_version=product_version,
        credentials_strategy=credentials_strategy,
        credentials_provider=credentials_provider,
        token_audience=token_audience,
        config=config,
        **os.environ,
    )


def get_dbutils(
    host: str | None = None,
    account_id: str | None = None,
    username: str | None = None,
    password: str | None = None,
    client_id: str | None = None,
    client_secret: str | None = None,
    token: str | None = None,
    profile: str | None = None,
    config_file: str | None = None,
    azure_workspace_resource_id: str | None = None,
    azure_client_secret: str | None = None,
    azure_client_id: str | None = None,
    azure_tenant_id: str | None = None,
    azure_environment: str | None = None,
    auth_type: str | None = None,
    cluster_id: str | None = None,
    google_credentials: str | None = None,
    google_service_account: str | None = None,
    debug_truncate_bytes: int | None = None,
    *,
    debug_headers: bool | None = None,
    product: str = "unknown",
    product_version: str = "0.0.0",
    credentials_strategy: CredentialsStrategy | None = None,
    credentials_provider: CredentialsStrategy | None = None,
    token_audience: str | None = None,
    config: Config | None = None,
) -> RemoteDbUtils:  # pragma: no cover - environment dependent
    """
    Get [dbutils](https://docs.databricks.com/dev-tools/databricks-utils.html)
    using an existing instance from the runtime if found, otherwise,
    creating one using a workspace client (this requires either having
    [these environment variables set
    ](https://docs.databricks.com/aws/en/dev-tools/auth#environment-variables),
    or providing the equivalent optional arguments.
    """
    dbutils: RemoteDbUtils | None = None
    with suppress(ImportError):
        from IPython.core.getipython import (  # type: ignore[import-not-found]  # noqa: PLC0415
            get_ipython,
        )

        if TYPE_CHECKING:
            from IPython.core.interactiveshell import (  # type: ignore[import-not-found]  # noqa: PLC0415
                InteractiveShell,
            )

        ipython: InteractiveShell = get_ipython()
        if ipython is not None:
            user_namespace_attribute: str
            user_namespace: dict
            for user_namespace_attribute in "user_ns", "user_global_ns":
                dbutils = getattr(ipython, user_namespace_attribute, {}).get(
                    "dbutils", None
                )
                if dbutils is not None:
                    return dbutils
    dbutils = globals().get("dbutils")
    if dbutils is not None:
        return dbutils
    databricks_workspace_client: WorkspaceClient = (
        get_databricks_workspace_client(
            host=host,
            account_id=account_id,
            username=username,
            password=password,
            client_id=client_id,
            client_secret=client_secret,
            token=token,
            profile=profile,
            config_file=config_file,
            azure_workspace_resource_id=azure_workspace_resource_id,
            azure_client_secret=azure_client_secret,
            azure_client_id=azure_client_id,
            azure_tenant_id=azure_tenant_id,
            azure_environment=azure_environment,
            auth_type=auth_type,
            cluster_id=cluster_id,
            google_credentials=google_credentials,
            google_service_account=google_service_account,
            debug_truncate_bytes=debug_truncate_bytes,
            debug_headers=debug_headers,
            product=product,
            product_version=product_version,
            credentials_strategy=credentials_strategy,
            credentials_provider=credentials_provider,
            token_audience=token_audience,
            config=config,
        )
    )
    return databricks_workspace_client.dbutils


@cache
def _get_secret(
    scope: str,
    key: str,
    host: str | None = None,
    account_id: str | None = None,
    username: str | None = None,
    password: str | None = None,
    client_id: str | None = None,
    client_secret: str | None = None,
    token: str | None = None,
    profile: str | None = None,
    config_file: str | None = None,
    azure_workspace_resource_id: str | None = None,
    azure_client_secret: str | None = None,
    azure_client_id: str | None = None,
    azure_tenant_id: str | None = None,
    azure_environment: str | None = None,
    auth_type: str | None = None,
    cluster_id: str | None = None,
    google_credentials: str | None = None,
    google_service_account: str | None = None,
    debug_truncate_bytes: int | None = None,
    *,
    debug_headers: bool | None = None,
    product: str = "unknown",
    product_version: str = "0.0.0",
    credentials_strategy: CredentialsStrategy | None = None,
    credentials_provider: CredentialsStrategy | None = None,
    token_audience: str | None = None,
    config: Config | None = None,
    **env: str,  # noqa: ARG001
) -> str:
    """
    Get a secret from Databricks, and cache it based on parameters and
    environment variables (since these can change the host and authentication).
    """
    return get_dbutils(
        host=host,
        account_id=account_id,
        username=username,
        password=password,
        client_id=client_id,
        client_secret=client_secret,
        token=token,
        profile=profile,
        config_file=config_file,
        azure_workspace_resource_id=azure_workspace_resource_id,
        azure_client_secret=azure_client_secret,
        azure_client_id=azure_client_id,
        azure_tenant_id=azure_tenant_id,
        azure_environment=azure_environment,
        auth_type=auth_type,
        cluster_id=cluster_id,
        google_credentials=google_credentials,
        google_service_account=google_service_account,
        debug_truncate_bytes=debug_truncate_bytes,
        debug_headers=debug_headers,
        product=product,
        product_version=product_version,
        credentials_strategy=credentials_strategy,
        credentials_provider=credentials_provider,
        token_audience=token_audience,
        config=config,
    ).secrets.get(scope, key)


def get_databricks_secret(
    scope: str,
    key: str,
    host: str | None = None,
    account_id: str | None = None,
    username: str | None = None,
    password: str | None = None,
    client_id: str | None = None,
    client_secret: str | None = None,
    token: str | None = None,
    profile: str | None = None,
    config_file: str | None = None,
    azure_workspace_resource_id: str | None = None,
    azure_client_secret: str | None = None,
    azure_client_id: str | None = None,
    azure_tenant_id: str | None = None,
    azure_environment: str | None = None,
    auth_type: str | None = None,
    cluster_id: str | None = None,
    google_credentials: str | None = None,
    google_service_account: str | None = None,
    debug_truncate_bytes: int | None = None,
    *,
    debug_headers: bool | None = None,
    product: str = "unknown",
    product_version: str = "0.0.0",
    credentials_strategy: CredentialsStrategy | None = None,
    credentials_provider: CredentialsStrategy | None = None,
    token_audience: str | None = None,
    config: Config | None = None,
) -> str:
    """
    Get a secret from Databricks.

    Parameters:
        scope: The Databricks secret scope.
        key: The Databricks secret key.
        host: A Databricks workspace host URL.
        account_id: A Databricks account ID.
        username: A Databricks username.
        password: A Databricks password.
        client_id: A Databricks OAuth2 Client ID.
        client_secret: A Databricks OAuth2 Client Secret.
        token: A Databricks Personal Access Token.
        profile: A Databricks Configuration Profile.
        config_file: A Databricks Configuration File path.
        azure_workspace_resource_id: An Azure Databricks Workspace Resource ID.
        azure_client_secret: An Azure Client Secret for Azure Databricks auth.
        azure_client_id: An Azure Client ID for Azure Databricks auth.
        azure_tenant_id: An Azure Tenant ID for Azure Databricks auth.
        azure_environment: An Azure Environment for Azure Databricks auth.
        auth_type: A Databricks authentication type.
        cluster_id: A Databricks cluster ID.
        google_credentials: Google Cloud credentials for GCP Databricks auth.
        google_service_account: A Google Service Account for GCP Databricks
            auth.
        debug_truncate_bytes: Number of bytes to truncate in debug logs.
        debug_headers: Whether to enable debug logging of HTTP headers.
        product: The product name using the SDK.
        product_version: The product version using the SDK.
        credentials_strategy: A credentials strategy for the SDK.
        credentials_provider: A credentials provider for the SDK.
        token_audience: A token audience for the SDK.
        config: A Databricks SDK Config instance.
    """
    return _get_secret(
        scope,
        key,
        host=host,
        account_id=account_id,
        username=username,
        password=password,
        client_id=client_id,
        client_secret=client_secret,
        token=token,
        profile=profile,
        config_file=config_file,
        azure_workspace_resource_id=azure_workspace_resource_id,
        azure_client_secret=azure_client_secret,
        azure_client_id=azure_client_id,
        azure_tenant_id=azure_tenant_id,
        azure_environment=azure_environment,
        auth_type=auth_type,
        cluster_id=cluster_id,
        google_credentials=google_credentials,
        google_service_account=google_service_account,
        debug_truncate_bytes=debug_truncate_bytes,
        debug_headers=debug_headers,
        product=product,
        product_version=product_version,
        credentials_strategy=credentials_strategy,
        credentials_provider=credentials_provider,
        token_audience=token_audience,
        config=config,
        **os.environ,
    )


def _get_scope_key_secret(
    scope_key: tuple[str, str] | str,
    host: str | None = None,
    account_id: str | None = None,
    username: str | None = None,
    password: str | None = None,
    client_id: str | None = None,
    client_secret: str | None = None,
    token: str | None = None,
    profile: str | None = None,
    config_file: str | None = None,
    azure_workspace_resource_id: str | None = None,
    azure_client_secret: str | None = None,
    azure_client_id: str | None = None,
    azure_tenant_id: str | None = None,
    azure_environment: str | None = None,
    auth_type: str | None = None,
    cluster_id: str | None = None,
    google_credentials: str | None = None,
    google_service_account: str | None = None,
    debug_truncate_bytes: int | None = None,
    *,
    debug_headers: bool | None = None,
    product: str = "unknown",
    product_version: str = "0.0.0",
    credentials_strategy: CredentialsStrategy | None = None,
    credentials_provider: CredentialsStrategy | None = None,
    token_audience: str | None = None,
    config: Config | None = None,
) -> str:
    if isinstance(scope_key, str):
        scope_key = scope_key.partition("/")[::2]
    return get_databricks_secret(
        *scope_key,
        host=host,
        account_id=account_id,
        username=username,
        password=password,
        client_id=client_id,
        client_secret=client_secret,
        token=token,
        profile=profile,
        config_file=config_file,
        azure_workspace_resource_id=azure_workspace_resource_id,
        azure_client_secret=azure_client_secret,
        azure_client_id=azure_client_id,
        azure_tenant_id=azure_tenant_id,
        azure_environment=azure_environment,
        auth_type=auth_type,
        cluster_id=cluster_id,
        google_credentials=google_credentials,
        google_service_account=google_service_account,
        debug_truncate_bytes=debug_truncate_bytes,
        debug_headers=debug_headers,
        product=product,
        product_version=product_version,
        credentials_strategy=credentials_strategy,
        credentials_provider=credentials_provider,
        token_audience=token_audience,
        config=config,
    )


def _get_args_options(
    *args: Any,
) -> tuple[tuple[Any, ...], DatabricksWorkspaceClientArguments | None]:
    """
    This function extracts an instance of `DatabricksWorkspaceConfigArguments`
    from the provided arguments, if one is present.
    """
    index: int
    value: Any
    for index, value in enumerate(args):
        if isinstance(value, DatabricksWorkspaceClientArguments):
            return (*args[:index], *args[index + 1 :]), value
    return args, None


def _print_help() -> None:
    print(  # noqa: T201
        "Usage:\n"
        "  decorative-secrets databricks <command> [options]\n\n"
        "Commands:\n"
        "  install\n"
        "  get"
    )


def _get_command() -> str:
    command: str = ""
    if len(sys.argv) > 1:
        command = sys.argv.pop(1).lower().replace("_", "-")
    return command


def main() -> None:
    """
    Run a command:
    -   install: Install the Databricks CLI if not already installed
    -   get: Get a secret from Databricks and print it to stdout
    """
    command = _get_command()
    if command in ("--help", "-h"):
        _print_help()
        return
    parser: argparse.ArgumentParser
    if command == "install":
        parser = argparse.ArgumentParser(
            prog="decorative-secrets databricks install",
            description="Install the Databricks CLI",
        )
        parser.parse_args()
        _install_databricks_cli()
    elif command == "get":
        parser = argparse.ArgumentParser(
            prog="decorative-secrets databricks get",
            description="Get a secret from Databricks",
        )
        parser.add_argument(
            "scope",
            type=str,
        )
        parser.add_argument(
            "key",
            type=str,
        )
        parser.add_argument(
            "--host",
            default=None,
            type=str,
            help="A Databricks workspace host URL",
        )
        parser.add_argument(
            "-cid",
            "--client-id",
            default=None,
            type=str,
            help="A Databricks OAuth2 Client ID",
        )
        parser.add_argument(
            "-cs",
            "--client-secret",
            default=None,
            type=str,
            help="A Databricks OAuth2 Client Secret",
        )
        parser.add_argument(
            "-t",
            "--token",
            default=None,
            type=str,
            help="A Databricks Personal Access Token",
        )
        parser.add_argument(
            "-p",
            "--profile",
            default=None,
            type=str,
            help="A Databricks Configuration Profile",
        )
        namespace: argparse.Namespace = parser.parse_args()
        print(  # noqa: T201
            get_databricks_secret(
                namespace.scope,
                namespace.key,
                host=namespace.host,
                client_id=namespace.client_id,
                client_secret=namespace.client_secret,
                token=namespace.token,
                profile=namespace.profile,
            )
        )


if __name__ == "__main__":
    main()



from __future__ import annotations

import argparse
import asyncio
import os
import sys
from contextlib import suppress
from dataclasses import dataclass
from functools import cache, partial
from importlib.metadata import distribution
from shutil import which
from subprocess import CalledProcessError
from typing import TYPE_CHECKING, Any
from urllib.parse import ParseResult, urlparse

from async_lru import alru_cache
from onepassword.client import Client  # type: ignore[import-untyped]
from onepasswordconnectsdk.client import (  # type: ignore[import-untyped]
    AsyncClient,
    Item,
)
from onepasswordconnectsdk.client import (  # type: ignore[import-untyped]
    Client as ConnectClient,
)

from decorative_secrets._utilities import (  # type: ignore[import-untyped]
    which_brew,
    which_winget,
)
from decorative_secrets.callback import apply_callback_arguments
from decorative_secrets.errors import (
    OnePasswordCommandLineInterfaceNotInstalledError,
    WinGetNotInstalledError,
)
from decorative_secrets.subprocess import check_output

if TYPE_CHECKING:
    from collections.abc import Callable, Coroutine, Iterable

    from onepassword import Secrets  # type: ignore[import-untyped]
    from onepasswordconnectsdk.models.field import (  # type: ignore[import-untyped]
        Field,
    )

_INTEGRATION_NAME: str = "decorative-secrets"
_INTEGRATION_VERSION: str = distribution("decorative-secrets").version


def apply_onepassword_arguments(
    *args: ApplyOnepasswordArgumentsOptions,
    **kwargs: str,
) -> Callable:
    """
    This decorator maps parameter names to 1Password resources.
    Each key represents the name of a parameter in the decorated function
    which accepts an explicit input, and the corresponding mapped value is a
    parameter name accepting a resource path with which to lookup a secret
    to pass to the mapped parameter in lieu of an explicitly provided
    argument.

    Parameters:
        *args: An optional [ApplyOnepasswordArgumentsOptions
            ](./#decorative_secrets.onepassword.ApplyOnepasswordArgumentsOptions)
            instance governing the behavior of this decorator. If not provided,
            a default instance of [ApplyOnepasswordArgumentsOptions()
            ](./#decorative_secrets.onepassword.ApplyOnepasswordArgumentsOptions)
            will be used. If multiple instances are provided, only the first
            will be used.
        **kwargs:
            A mapping of static parameter names to the parameter names
            of arguments accepting 1Password resource paths from which to
            retrieve the value when the key argument is not explicitly
            provided.

    Example:
        ```python
        from functools import (
            cache,
        )
        from decorative_secrets.onepassword import (
            apply_onepassword_arguments,
        )
        from my_client_sdk import (
            Client,
        )


        @cache
        @apply_onepassword_arguments(
            client_id="client_id_onepassword",
            client_secret="client_secret_onepassword",
        )
        def get_client(
            client_id: str | None = None,
            client_secret: str = None,
            client_id_onepassword: str | None = None,
            client_secret_onepassword: str | None = None,
        ) -> Client:
            return Client(
                oauth2_client_id=client_id,
                oauth2_client_secret=client_secret,
            )


        client: Client = get_client(
            client_id_onepassword=(
                "op://Vault Name/Client ID Item Name/username",
            ),
            client_secret_onepassword=(
                "op://Vault Name/Client Secret Item Name/credential",
            ),
        )
        ```
    """
    options: ApplyOnepasswordArgumentsOptions
    args, options = _get_args_options(*args)
    read_onepassword_secret_: Callable[..., str] = read_onepassword_secret
    async_read_onepassword_secret_: Callable[
        [str, str | None, str | None, str | None], Coroutine[Any, Any, str]
    ] = async_read_onepassword_secret
    if (
        (options.account is not None)
        or (options.token is not None)
        or (options.host is not None)
    ):
        read_onepassword_secret_ = partial(
            read_onepassword_secret_,
            **({"account": options.account} if options.account else {}),
            **({"token": options.token} if options.token else {}),
            **({"host": options.host} if options.host else {}),
        )
        async_read_onepassword_secret_ = partial(
            async_read_onepassword_secret_,
            **({"account": options.account} if options.account else {}),
            **({"token": options.token} if options.token else {}),
            **({"host": options.host} if options.host else {}),
        )
    return apply_callback_arguments(
        read_onepassword_secret_,
        async_read_onepassword_secret_,
        **kwargs,
    )


def _install_op() -> None:
    """
    Install the 1Password CLI.
    """
    message: str
    if sys.platform.startswith("win"):
        try:
            check_output((which_winget(), "install", "1password-cli"))
        except (
            CalledProcessError,
            FileNotFoundError,
            WinGetNotInstalledError,
        ) as error:
            raise OnePasswordCommandLineInterfaceNotInstalledError from error
    elif sys.platform == "darwin":
        try:
            check_output((which_brew(), "install", "1password-cli"))
        except (CalledProcessError, FileNotFoundError) as error:
            raise OnePasswordCommandLineInterfaceNotInstalledError from error
    else:
        raise OnePasswordCommandLineInterfaceNotInstalledError


def which_op() -> str:
    """
    Locate the 1Password CLI executable, or attempt
    to install it if not found.
    """
    op: str = which("op") or "op"
    try:
        check_output((op, "--version"))
    except (CalledProcessError, FileNotFoundError):
        _install_op()
        op = which("op") or "op"
    return op


@cache
def _op_signin(account: str | None = None) -> str:
    op: str = which_op()
    if not account:
        account = os.getenv("OP_ACCOUNT")
    check_output(
        (op, "signin", "--account", account) if account else (op, "signin"),
        input=None if account else b"\n\n",
    )
    return op


def iter_op_account_list() -> Iterable[str]:
    """
    Yield all 1password account names.
    """
    op: str = which_op()
    line: str
    for line in check_output((op, "account", "list")).strip().split("\n")[1:]:
        yield line.partition(" ")[0]


def op_signin(account: str | None = None) -> str:
    """
    Sign in to 1Password using the CLI if not already signed in.
    """
    account = account or os.getenv("OP_ACCOUNT")
    if account:
        return _op_signin(account)
    op: str | None = None
    for account in iter_op_account_list():
        op = _op_signin(account)
    return op or which_op()


def _resolve_auth_arguments(
    account: str | None = None,
    token: str | None = None,
    host: str | None = None,
) -> tuple[str | None, str | None, str | None]:
    """
    Parameters:
        account:
        token:
        host:
    """
    if account is None:
        account = os.getenv("OP_ACCOUNT")
    if token is None:
        token = os.getenv("OP_SERVICE_ACCOUNT_TOKEN")
    if host is None:
        host = os.getenv("OP_CONNECT_HOST")
    if (token is None) or (host is not None):
        token = os.getenv("OP_CONNECT_TOKEN") or token
    return account, token, host


def _parse_resource(resource: str) -> tuple[str, str, str]:
    parse_result: ParseResult = urlparse(resource)
    return (parse_result.netloc, *parse_result.path[1:].partition("/")[::2])


async def _async_resolve_resource(token: str, resource: str) -> str:
    """
    Asynchronously resolve a 1Password resource using the
    `onepassword-sdk` library.
    """
    client: Client = await Client.authenticate(
        auth=token,
        integration_name=_INTEGRATION_NAME,
        integration_version=_INTEGRATION_VERSION,
    )
    secrets: Secrets = client.secrets
    return await secrets.resolve(resource)


async def _async_resolve_connect_resource(
    token: str, host: str, resource: str
) -> str:  # pragma: no cover
    connect_client: AsyncClient = AsyncClient(
        url=host,
        token=token,
    )
    vault: str
    item_name: str
    field_id: str
    vault, item_name, field_id = _parse_resource(resource)
    item: Item = await connect_client.get_item(item=item_name, vault=vault)
    field: Field
    for field in item.fields:
        if field.id == field_id:
            return field.value
    raise KeyError(resource)


def _resolve_connect_resource(
    token: str, host: str, resource: str
) -> str:  # pragma: no cover
    connect_client: ConnectClient = ConnectClient(
        url=host,
        token=token,
    )
    vault: str
    item_name: str
    field_id: str
    vault, item_name, field_id = _parse_resource(resource)
    item: Item = connect_client.get_item(item=item_name, vault=vault)
    field: Field
    for field in item.fields:
        if field.id == field_id:
            return field.value
    raise KeyError(resource)


@alru_cache(maxsize=None)
async def async_read_onepassword_secret(
    resource: str,
    account: str | None = None,
    token: str | None = None,
    host: str | None = None,
) -> str:
    """
    Asynchronously read a secret from 1Password using either the
    `onepassword-sdk` or `onepasswordconnectsdk` libraries, or the `op`
    executable (1password CLI), depending on the provided arguments and
    environment variables.

    Parameters:
        resource: A 1Password secret resource path. For example:
            "op://Vault Name/Client Secret Item Name/credential"
        account: A 1Password account URL. For example, individuals and families
            will use "my.1password.com", while teams and businesses will use
            a custom subdomain. This is only necessary when using
            the 1Password CLI where multiple accounts are configured.
        token: A 1Password or 1Password connect service account token.
        host: A 1Password Connect host URL. This is required when using
            self-hosted 1Password Connect.

    Returns:
        The resolved secret value.
    """
    account, token, host = _resolve_auth_arguments(account, token, host)
    if token:
        if host:
            return await _async_resolve_connect_resource(token, host, resource)
        return await _async_resolve_resource(token, resource)
    op: str | None = None
    with suppress(FileNotFoundError, CalledProcessError):
        op = op_signin(account)
    if not op:
        op = which_op() or "op"
    return check_output(
        (op, "read")
        + (("--account", account) if account else ())
        + (("--session", token) if token else ())
        + (resource,)
    )


@cache
def _read_onepassword_secret(
    resource: str,
    account: str | None = None,
    token: str | None = None,
    host: str | None = None,
    **env: str,  # noqa: ARG001
) -> str:
    """
    This function is wrapped by `read_onepassword_secret` to allow caching
    to be invalidated based on environment variable changes.
    """
    account, token, host = _resolve_auth_arguments(account, token, host)
    if token:
        if host:
            return _resolve_connect_resource(token, host, resource)
        return asyncio.run(_async_resolve_resource(token, resource))
    op: str | None = None
    with suppress(FileNotFoundError, CalledProcessError):
        op = op_signin(account)
    if not op:
        op = which_op() or "op"
    return check_output(
        (op, "read")
        + (("--account", account) if account else ())
        + (("--session", token) if token else ())
        + (resource,)
    )


def get_onepassword_secret(
    resource: str,
    account: str | None = None,
    token: str | None = None,
    host: str | None = None,
) -> str:
    """
    Read a secret from 1Password using either the `onepassword-sdk` or
    `onepasswordconnectsdk` libraries, or the `op` executable (1password CLI),
    depending on the provided arguments and environment variables.

    Parameters:
        resource: A 1Password secret resource path. For example:
            "op://Vault Name/Client Secret Item Name/credential"
        account: A 1Password account URL. For example, individuals and families
            will use "my.1password.com", while teams and businesses will use
            a custom subdomain. This is only necessary when using
            the 1Password CLI where multiple accounts are configured.
        token: A 1Password or 1Password connect service account token.
        host: A 1Password Connect host URL. This is required when using
            self-hosted 1Password Connect.

    Returns:
        The resolved secret value.
    """
    return _read_onepassword_secret(
        resource, account=account, token=token, host=host, **os.environ
    )


# For backward compatibility
read_onepassword_secret = get_onepassword_secret  # type: ignore[assignment]


@dataclass(frozen=True)
class ApplyOnepasswordArgumentsOptions:
    """
    This class contains options governing the behavior of the
    [apply_onepassword_arguments
    ](./#decorative_secrets.onepassword.apply_onepassword_arguments) decorator.

    Attributes:
        account: A 1Password account URL. For example, individuals
            and families will use "my.1password.com", while teams and
            businesses will use a custom subdomain. If not provided, the
            `OP_ACCOUNT` environment variable will be used, if set. This is
            only necessary when using the 1Password CLI where multiple
            accounts are configured, and if no token is provided or inferred
            from an environment variable.
        token: A 1Password or 1Password connect service account
            token. If not provided, the `OP_SERVICE_ACCOUNT_TOKEN` or
            `OP_CONNECT_TOKEN` environment variables will be used, if set.
        host: A 1Password Connect host URL. If not
            provided, the `OP_CONNECT_HOST` environment variable will be used,
            if set. This is required when using a self-hosted 1Password
            Connect server.
    """

    account: str | None = None
    token: str | None = None
    host: str | None = None


def _get_args_options(
    *args: Any,
) -> tuple[tuple[Any, ...], ApplyOnepasswordArgumentsOptions]:
    """
    This function extracts an `ApplyEnvironmentArgumentsOptions` instance
    from the provided arguments, if one is present.
    """
    index: int
    value: Any
    for index, value in enumerate(args):
        if isinstance(value, ApplyOnepasswordArgumentsOptions):
            return (*args[:index], *args[index + 1 :]), value
    return args, ApplyOnepasswordArgumentsOptions()


def _print_help() -> None:
    print(  # noqa: T201
        "Usage:\n"
        "  decorative-secrets onepassword <command> [options]\n\n"
        "Commands:\n"
        "  install\n"
        "  get"
    )


def _get_command() -> str:
    command: str = ""
    if len(sys.argv) > 1:
        command = sys.argv.pop(1).lower().replace("_", "-")
    return command


def main() -> None:
    """
    Run a command:
    -   install: Install the Databricks CLI if not already installed
    -   get: Get a secret from Databricks and print it to stdout
    """
    command = _get_command()
    if command in ("--help", "-h"):
        _print_help()
        return
    parser: argparse.ArgumentParser
    if command == "install":
        parser = argparse.ArgumentParser(
            prog="decorative-secrets onepassword install",
            description="Install the 1Password CLI",
        )
        parser.parse_args()
        _install_op()
    elif command == "get":
        parser = argparse.ArgumentParser(
            prog="decorative-secrets onepassword get",
            description="Get a secret from 1Password",
        )
        parser.add_argument(
            "reference",
            type=str,
        )
        parser.add_argument(
            "--account",
            default=None,
            type=str,
            help="Which 1Password account to use",
        )
        parser.add_argument(
            "-t",
            "--token",
            default=None,
            type=str,
            help="A 1Password Service Account Token",
        )
        parser.add_argument(
            "--host",
            default=None,
            type=str,
            help="A 1Password Connect Host URL",
        )
        namespace: argparse.Namespace = parser.parse_args()
        print(  # noqa: T201
            read_onepassword_secret(
                namespace.reference,
                host=namespace.host,
                account=namespace.account,
                token=namespace.token,
            )
        )


if __name__ == "__main__":
    main()



from __future__ import annotations

from io import StringIO
from subprocess import (
    PIPE,
    CalledProcessError,
    CompletedProcess,
    list2cmdline,
    run,
)
from tempfile import TemporaryFile
from typing import TYPE_CHECKING, Literal, overload

if TYPE_CHECKING:
    from collections.abc import Mapping
    from pathlib import Path


@overload
def check_output(
    args: tuple[str, ...],
    *,
    text: Literal[True] = True,
    cwd: str | Path | None = None,
    input: str | bytes | None = None,
    env: Mapping[str, str] | None = None,
    echo: bool = False,
) -> str: ...


@overload
def check_output(
    args: tuple[str, ...],
    *,
    text: Literal[False] = False,
    cwd: str | Path | None = None,
    input: str | bytes | None = None,
    env: Mapping[str, str] | None = None,
    suppress_stderr: bool = True,
    echo: bool = False,
) -> bytes: ...


@overload
def check_output(
    args: tuple[str, ...],
    *,
    text: None = None,
    cwd: str | Path | None = None,
    input: str | bytes | None = None,
    env: Mapping[str, str] | None = None,
    suppress_stderr: bool = True,
    echo: bool = False,
) -> bytes: ...


def check_output(  # noqa: C901
    args: tuple[str, ...],
    *,
    text: bool | None = True,
    cwd: str | Path | None = None,
    input: str | bytes | None = None,  # noqa: A002
    env: Mapping[str, str] | None = None,
    suppress_stderr: bool = True,
    echo: bool = False,
) -> str | bytes | None:
    """
    This function mimics `subprocess.check_output`, but redirects stderr
    to DEVNULL, ignores unicode decoding errors, and outputs text by default.

    Parameters:
        args: The command to run
        text: Whether to return output as text (default: `True`). If
            `None`—returns `None`. If `False`, returns the output as
            `bytes`.
        cwd: The working directory to run the command in
        input: Input to send to the command
        env: Environment variables to set for the command
        echo: Whether to print the command and its output (default: False)
        suppress_stderr: Whether to prevent stderr from being printed to the
            console
    """
    if echo:
        if cwd:
            print("$", "cd", cwd, "&&", list2cmdline(args))  # noqa: T201
        else:
            print("$", list2cmdline(args))  # noqa: T201
    if isinstance(input, bytes) and text:
        input = input.decode("utf-8", errors="ignore")  # noqa: A001

    completed_process: CompletedProcess
    if suppress_stderr:
        with TemporaryFile("w+") as stderr:
            try:
                completed_process = run(
                    args,
                    stdout=PIPE,
                    stderr=stderr,  # DEVNULL,
                    check=True,
                    cwd=cwd or None,
                    input=input,
                    env=env,
                    text=text,
                )
            except CalledProcessError as error:
                error.stderr = StringIO(stderr.read())
                raise
    else:
        completed_process = run(
            args,
            capture_output=True,
            check=True,
            cwd=cwd or None,
            input=input,
            env=env,
            text=text,
        )
    output: str | bytes | None = None
    if text is None:
        pass
    elif text:
        output = completed_process.stdout.rstrip()
        if isinstance(output, bytes):
            output = output.decode("utf-8", errors="ignore")
    else:
        output = completed_process.stdout.rstrip()
        if isinstance(output, str):
            output = output.encode("utf-8", errors="ignore")
    if echo and (output is not None):
        print(output)  # noqa: T201
    return output


def check_call(
    args: tuple[str, ...],
    *,
    cwd: str | Path | None = None,
    input: str | bytes | None = None,  # noqa: A002
    env: Mapping[str, str] | None = None,
    suppress_stderr: bool = True,
    echo: bool = False,
) -> None:
    """
    This function mimics `subprocess.check_call`, but redirects stderr
    to DEVNULL.

    Parameters:
        args: The command to run
        text: Whether to return output as text (default: `True`). If
            `None`—returns `None`. If `False`, returns the output as
            `bytes`.
        cwd: The working directory to run the command in
        input: Input to send to the command
        env: Environment variables to set for the command
        echo: Whether to print the command and its output (default: False)
        suppress_stderr: Whether to prevent stderr from being printed to the
            console
    """
    check_output(
        args,
        text=None,
        cwd=cwd,
        input=input,
        env=env,
        suppress_stderr=suppress_stderr,
        echo=echo,
    )
